#!/bin/sh
# SPDX-License-Identifier: BSD-3-Clause
# Copyright Contributors to the OpenColorIO Project.

# For OS X
export DYLD_LIBRARY_PATH="/usr/local/lib:${DYLD_LIBRARY_PATH}"

# For Linux
export LD_LIBRARY_PATH="/usr/local/lib:${LD_LIBRARY_PATH}"

export PATH="/usr/local/bin:${PATH}"
export PYTHONPATH="/usr/local/lib/python3.13/site-packages:${PYTHONPATH}"
